import logging
from inspect import currentframe
from lytils import ctext
from lytils.file import LyFile


def get_line_number():
    cf = currentframe()
    return cf.f_back.f_lineno


class InvalidLogLevelException(Exception):
    # Raise this when undetected_chromedriver is not installed
    def __init__(
        self,
        message=f"<y>Parameter 'level' must equal one of the following: debug, info, warning, error, critical.",
    ):
        self.message = ctext(message)
        super().__init__(self.message)


class LyLoggerNotSetUpException(Exception):
    # Raise this when undetected_chromedriver is not installed
    def __init__(
        self,
        message=f"<y>LyLogger.setup() not ran.",
    ):
        self.message = ctext(message)
        super().__init__(self.message)


class LyLogger:
    LEVELS = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }

    def __init__(
        self,
        name: str,
        path: str,
        encoding: str = "utf-8",
        formatting: str = "%(asctime)s - [%(levelname)s] %(message)s",
        date_format: str = "%Y-%m-%d %H:%M:%S",
        level="debug",
        debug_path: str = "",
        info_path: str = "",
        warning_path: str = "",
        error_path: str = "",
        critical_path: str = "",
    ):
        self._name = name
        self._path = path
        self._encoding = encoding
        self._formatting = formatting
        self._date_format = date_format

        if level not in LyLogger.LEVELS.keys():
            raise InvalidLogLevelException
        self._level = LyLogger.LEVELS[level]

        # Allow for multiple log files
        self._debug_path = debug_path if debug_path else path
        self._info_path = info_path if info_path else path
        self._warning_path = warning_path if warning_path else path
        self._error_path = error_path if error_path else path
        self._critical_path = critical_path if critical_path else path

        self._logger = None

    def _get_file_handler(self, path: str, level, formatter):
        file_handler = logging.FileHandler(path)
        file_handler.setLevel(LyLogger.LEVELS[level])
        file_handler.setFormatter(formatter)
        return file_handler

    def setup(self, overwrite: bool = False):
        log_file = LyFile(self._path)

        # Create file if it doesn't exist or we wish to overwrite the log each run.
        if overwrite:
            log_file.create()
        elif not log_file.exists():
            log_file.create()

        logger = logging.getLogger(self._name)
        logger.setLevel(self._level)

        formatter = logging.Formatter(self._formatting)

        # Get and add file handlers
        debug_handler = self._get_file_handler(self._debug_path, "debug", formatter)
        logger.addHandler(debug_handler)

        info_handler = self._get_file_handler(self._info_path, "info", formatter)
        logger.addHandler(info_handler)

        warning_handler = self._get_file_handler(
            self._warning_path, "warning", formatter
        )
        logger.addHandler(warning_handler)

        error_handler = self._get_file_handler(self._error_path, "error", formatter)
        logger.addHandler(error_handler)

        critical_handler = self._get_file_handler(
            self._critical_path, "critical", formatter
        )
        logger.addHandler(critical_handler)

        self._logger = logger

    def debug(self, message: str):
        if not self._logger:
            raise LyLoggerNotSetUpException
        self._logger.debug(message)

    def info(self, message: str):
        if not self._logger:
            raise LyLoggerNotSetUpException
        self._logger.info(message)

    def warning(self, message: str):
        if not self._logger:
            raise LyLoggerNotSetUpException
        self._logger.warning(message)

    def error(self, message: str):
        if not self._logger:
            raise LyLoggerNotSetUpException
        self._logger.error(message)

    def critical(self, message: str):
        if not self._logger:
            raise LyLoggerNotSetUpException
        self._logger.critical(message)

    def path(self):
        return self._path

    def debug_path(self):
        return self._debug_path

    def info_path(self):
        return self._info_path

    def warning_path(self):
        return self._warning_path

    def error_path(self):
        return self._error_path

    def critical_path(self):
        return self._critical_path



# Third-party Libraries
import smtplib
from email.message import EmailMessage
from lytils.surfshark import Surfshark


class Email:
    def __init__(self, smtp, credentials, surfshark_path: str = "", notify: str = ""):
        """
        :param smtp dict:
            Expects 'server' key (ex. 'smtp.example.com')
            Expects 'port' key (default: 587)
        :param credentials dict:
            Expects 'email' key; email used to send the email
            Expects 'password' key; password for sender's email
        :param notify string:
            Instead of having to pass a recipient email to
            send_email every time, you can declare a default
            email, and use notify() to send email to that recipient
        """
        self._smtp = smtp

        # Set port by default if not specified
        if "port" not in smtp:
            self._smtp["port"] = 587

        self._surfshark = None
        if surfshark_path:
            self._surfshark = Surfshark(surfshark_path)

        self._creds = credentials
        self._notify = notify
        self._server = None

    def start_server(self):
        # Kill surfshark if it is running as it blocks SMTP connections.
        if self._surfshark and self._surfshark.running():
            self._surfshark.kill()

        # Connect to the SMTP server
        self._server = smtplib.SMTP("smtp.gmail.com", 587)
        # self._server = smtplib.SMTP(smtp['server'], smtp['port'])
        # self._server.ehlo()
        self._server.starttls()
        self._server.login(self._creds["email"], self._creds["password"])

    def send_email(self, recipient, subject, message):
        # Create the email
        msg = EmailMessage()
        msg.set_content(message)
        msg["subject"] = subject
        msg["to"] = recipient
        msg["from"] = self._creds["email"]

        # Send the email
        self._server.sendmail(self._creds["email"], recipient, msg.as_string())

    def notify(self, subject, message):
        if self._notify:
            self.send_email(self._notify, subject, message)

    def quit(self):
        if self._server:
            self._server.quit()

        # Restart surfshark if it is not currently running.
        if self._surfshark and not self._surfshark.running():
            self._surfshark.start()



# Third-party libraries
from gspread import Worksheet
from gspread_formatting import batch_updater
from gspread_formatting import format_cell_range
from pandas import DataFrame

# Local libraries
from lytils import cprint
from lytils.google_sheets.format import HeaderFormat
from lytils.regex import match
from .Column import Column


# Get column letter from column number.
# 1 = A, 2 = B, ...,  27 = AA, 28 = AB, etc
def get_column_letter(column_number):
    dividend = column_number
    letter = ""
    while dividend > 0:
        remainder = (dividend - 1) % 26
        letter = chr(65 + remainder) + letter
        dividend = (dividend - remainder) // 26
    return letter


# Get column number from column letter.
# A = 1, B = 2, ...,  AA = 27, AB = 28, etc
def get_column_number(column_letter):
    number = 0
    for char in column_letter:
        number = number * 26 + (ord(char) - ord("A") + 1)
    return number


# Get header range from a range in a1 notation
def get_header_range(range: str):
    first_column = match(r"[A-Z]+", range, group=1)
    last_column = match(r"[A-Z]+", range, group=2)
    first_row = match(r"[0-9]+", range, group=1)

    return f"{first_column}{first_row}:{last_column}{first_row}"


# Get data range from a range in a1 notation
def get_data_range(range: str):
    first_column = match(r"[A-Z]+", range, group=1)
    first_row = match(r"[0-9]+", range, group=1)
    last_column = match(r"[A-Z]+", range, group=2)
    last_row = match(r"[0-9]+", range, group=2)

    # Increment first row by 1
    return f"{first_column}{int(first_row) + 1}:{last_column}{last_row}"


class Tab:
    def __init__(self, tab: Worksheet):
        self.__tab = tab
        self.__batch = batch_updater(self.__tab.spreadsheet)

    # Get path in terms of Spreadsheet > Worksheet
    def get_path(self):
        return f"{self.__tab.spreadsheet.title} > {self.__tab.title}"

    # Get all data
    def get_all_data(self) -> DataFrame:
        data = self.__tab.get_values()
        if len(data) > 1:
            return DataFrame(data[1:], columns=data[0])
        else:
            return DataFrame()

    # Get data from specific range
    def get_range(self, range: str, headers: bool = True) -> DataFrame:
        data = self.__tab.get(range)

        if data:
            if headers:
                headers_row = data.pop(0)  # Use the first row as headers
                df = DataFrame(data, columns=headers_row)
            else:
                df = DataFrame(data)
        else:
            df = DataFrame()

        df = df.fillna("")  # Replace None values with an empty string
        return df

    # Clear all data
    def clear(self) -> None:
        self.__tab.clear()

    # Clear specific ranges
    def clear_ranges(self, ranges: list[str]) -> None:
        self.__tab.batch_clear(ranges=ranges)

    # Format single column
    def format_column(self, header, format, header_row=1):
        headers = self.__tab.row_values(header_row)
        header_index = (
            headers.index(header) + 1
        )  # +1 because Google Sheets is 1-indexed
        column_letter = get_column_letter(header_index)
        format_range = f"{column_letter}{header_row + 1}:{column_letter}"
        format_cell_range(self.__tab, format_range, format)

    # Apply formatting to columns
    def format_columns(
        self,
        columns: list[Column],
        column_start: str = "A",
        row_start: int = 2,  # Row starts at 2 to account for headers
    ):
        col_start = get_column_number(column_start)
        formats = []
        for i, col in enumerate(columns, start=col_start):
            column_letter = get_column_letter(i)
            formats.append(
                [
                    f"{column_letter}{row_start}:{column_letter}",
                    col.get_format(),
                ]
            )
        self.__batch.format_cell_ranges(self.__tab, formats)

    # Set desired column widths
    def set_column_widths(self, columns: list[Column], column_start: int = 1):
        widths = []
        for i, col in enumerate(columns, start=column_start):
            column_letter = get_column_letter(i)
            widths.append(
                [
                    f"{column_letter}",
                    col.get_width(),
                ]
            )
        self.__batch.set_column_widths(self.__tab, widths)

    # Clear worksheet and populate with new data. Assumes headers.
    def set_all_data(self, data_frame: DataFrame, columns: list[Column]) -> None:
        self.clear()

        cprint(f"Outputting data to <c>{self.get_path()}<w>...")

        # Correct column order
        order = [col.get_header() for col in columns]
        ordered_df = data_frame[order]

        # Replace NaN with empty string, resulting in empty cells
        df = ordered_df.fillna(value="")

        headers = df.columns.tolist()
        data = df.values.tolist()

        # Format headers
        self.__batch.format_cell_range(self.__tab, "1:1", HeaderFormat())

        # * Format after setting data to include new rows in formatting
        self.format_columns(columns)
        self.set_column_widths(columns)

        # Execute bulk request
        self.__batch.execute()

        # * Formatting has to come BEFORE updates to format correctly

        # Set headers
        self.__tab.update("1:1", [headers])

        # Freeze the top row
        self.__tab.freeze(rows=1)

        # Set data
        self.__tab.update(
            range_name="A2",
            values=data,
            value_input_option="USER_ENTERED",
        )

    # Clear range and set new range with data. Assumes headers.
    def set_range(
        self,
        range: str,
        data_frame: DataFrame,
        columns: list[Column],
    ) -> None:
        cprint(f"Outputting data to <c>{self.get_path()} ({range})<w>...")

        # Correct column order
        order = [col.get_header() for col in columns]
        try:
            ordered_df = data_frame[order]
        except KeyError:
            cprint(f"<y>df columns: {data_frame.columns}")
            cprint(f"<y>order: {order}")
            raise KeyError

        # Replace NaN with empty string, resulting in empty cells
        df = ordered_df.fillna(value="")

        header_range = get_header_range(range)
        data_range = get_data_range(range)

        headers = df.columns.tolist()
        data = df.values.tolist()

        self.clear_ranges(ranges=[range])

        # Format headers
        self.__batch.format_cell_range(self.__tab, header_range, HeaderFormat())

        # * Format after setting data to include new rows in formatting
        column_start = match(r"[A-Z]+", data_range, group=1)
        row_start = match(r"[0-9]+", data_range, group=1)
        self.format_columns(columns, column_start=column_start, row_start=row_start)
        self.set_column_widths(columns, column_start=get_column_number(column_start))

        # Execute bulk request
        self.__batch.execute()

        # * Formatting has to come BEFORE updates to format correctly

        # Set headers
        self.__tab.update(header_range, [headers])

        # Set data
        self.__tab.update(
            range_name=data_range,
            values=data,
            value_input_option="USER_ENTERED",
        )



import traceback

COLORS = {
    "white": '\033[0m', # default,
    "red": '\033[31m',
    "green": '\033[32m',
    "yellow": '\033[33m',
    "blue": '\033[34m',
    "magenta": '\033[35m',
    "cyan": '\033[36m'
}

def ctext(text):
    text = text.replace("<w>", COLORS['white'])
    text = text.replace("<r>", COLORS['red'])
    text = text.replace("<g>", COLORS['green'])
    text = text.replace("<y>", COLORS['yellow'])
    text = text.replace("<b>", COLORS['blue'])
    text = text.replace("<m>", COLORS['magenta'])
    text = text.replace("<c>", COLORS['cyan'])
    
    return f"{text}{COLORS['white']}"

def cprint(text, level = 0):
    text = ctext(text)

    str_level = ""
    for i in range(0, level):
        str_level += "  "
    
    print(f"{str_level}{text}{COLORS['white']}")

def print_trace(color='red'):
    set_print_color(color)
    traceback.print_exc()
    reset_print_color()

def reset_print_color():
    print(COLORS['white'])

def set_print_color(color):
    print(COLORS[color])


from gspread import Spreadsheet
from .Tab import Tab


class Sheet:
    def __init__(self, sheet: Spreadsheet):
        self.__sheet = sheet

    def get_tab(self, name: str) -> Tab:
        tab = self.__sheet.worksheet(name)
        return Tab(tab=tab)



# Third-party Libraries
import gspread as gs

# Local Libraries
from lytils.google_sheets.Sheet import Sheet


# src: https://practicaldatascience.co.uk/data-science/how-to-read-google-sheets-data-in-pandas-with-gspread
class GoogleSheets:
    # Interface with GoogleSheet API to transform data

    def __init__(self, service_account: str):
        # Authenticate with Google Sheets
        self.__service_account = gs.service_account(filename=service_account)

    def get_sheet(self, url) -> Sheet:
        sheet = self.__service_account.open_by_url(url)
        return Sheet(sheet=sheet)



import discord
from lytils import cprint


class DiscordBot:

    def __init__(self, token: str, intents, debug: bool = False):
        """
        @token : string - token representing current bot generated by Discord
        @intents : Intents - object representing Discord permissions
            message_content set to False by default
        @debug : boolean - whether to display debug statements or not
        """
        self.set_client(intents)
        self.__token = token
        self.__debug = debug
        self.on_ready = None
        self.on_message = None

    def set_client(self, intents):
        self.__client = discord.Client(intents=intents)
    
    def get_client(self):
        return self.__client
    
    def get_channel(self, server_id, channel_name: str):
        # Get server by id
        server = discord.utils.get(self.__client.guilds, id=server_id)

        if server:
            # Find the channel by name
            channel = discord.utils.get(server.channels, name=channel_name)

            if channel:
                return channel
            else:
                cprint(f'<y>Channel \'{channel_name}\' not found in \'{server.name}\'.')
        else:
            cprint(f'<y>Server Id \'{server_id}\' not found.')

        return None


    def set_on_ready(self):
        # To customize on_ready method, set bot.on_ready to async def
        @self.__client.event
        async def on_ready():
            cprint(f'<g>{self.__client.user} is now running!')
            if self.on_ready is not None:
                await self.on_ready()

    def set_on_message(self):
        # To customize on_message method, set bot.on_message to async def
        @self.__client.event
        async def on_message(message):
            if message.author == self.__client.user:
                # This is to prevent the bot from responding to itself.
                return

            user = str(message.author)
            user_message = str(message.content)
            channel = str(message.channel)
            if self.__debug:
                cprint(f'<y>{user} said: "{user_message}" ({channel})')

            if self.on_message is not None:
                await self.on_message(message)
            else:
                if user_message[0] == '?':
                    # '?' is default for requesting a private response I guess.
                    user_message = user_message[1:]
                    await self.send_private_message(message, user_message)
                else:
                    await self.send_message(message, user_message)

    def get_response(message: str) -> str:
        # To customize get_response method, set bot.get_response to func
        # Lowercase message to handle case insensitive text
        p_message = message.lower()

        if p_message == 'hello':
            return 'Hey there!'

        if p_message == '!help':
            return '`This is a help message that you can modify.`'

    async def send_message(self, message, user_message: str):
        # Await this within your custom on_message function
        # Respond to channel which message came from
        try:
            response = DiscordBot.get_response(user_message)
            await message.channel.send(response)
        except Exception as e:
            cprint(f'<r>Error: {e}')

    async def send_private_message(self, message, user_message: str):
        # Await this within your custom on_message function
        # Respond privately to sender which message came from
        try:
            response = DiscordBot.get_response(user_message)
            await message.author.send(response)
        except Exception as e:
            cprint(f'<r>Error: {e}')

    async def send_message_to_channel(
            self,
            server_id,
            channel_name: str,
            message: str
        ):
        # Await this within your custom on_message function
        # Respond to channel which message came from
        try:
            # Get the 'norcal' text channel in the current server
            channel = self.get_channel(server_id, channel_name)
            await channel.send(message)
        except Exception as e:
            cprint(f'<r>Error: {e}')

    def run(self):
        self.set_on_ready()
        self.set_on_message()
        self.__client.run(self.__token)

    async def close(self):
        await self.__client.close()



from re import finditer, search, sub


# @param: group (int): Get first group, second group, etc. 1-based.
def match(regex: str, text: str, group: int = 1) -> str:
    matches = list(finditer(regex, text))
    try:
        return matches[group - 1].group(0)
    except IndexError:
        return ""


def replace(regex: str, replace_with: str, text: str) -> str:
    return sub(regex, replace_with, text)



import csv
import json
import os


def get_last_id_in_csv_file(file_name, column="id"):
    with open(file_name, "r") as file:
        reader = csv.DictReader(file)

        # Iterate over each row in the CSV file
        for row in reader:
            # Assign the last row to the last_row variable
            last_row = row

        try:
            # Get the value you want from the last row by column name
            desired_value = last_row[column]

            return int(desired_value)
        except:
            return -1


def load_json_from_file(path):
    """
    Load json object from file.
    """
    with open(path, "r") as file:
        return json.load(file)


def write_json_to_file(path, data, indent: int = 4):
    """
    Write json object to file.
    """
    with open(path, "w") as file:
        json.dump(data, file, indent=indent)


class LyFile:
    def __init__(self, path: str = "LyFile/file.txt"):
        self._path = path

    def exists(self):
        return os.path.exists(self._path)

    def create(self):
        """
        Creates a blank file at path.
        """
        # Split path into directory and filename
        directory, _ = os.path.split(self._path)

        # If directory was included, create it if it doesn't exist
        if directory and not os.path.exists(directory):
            os.makedirs(directory)

        open(self._path, "w").close()  # Create file

    def append(self, text):
        """
        Appends text to file at path.
        """
        with open(self._path, "a") as file:
            file.write(f"{text}\n")

    def append_json(self, data, indent: int = 4):
        """
        Appends json data to file at path.
        """
        with open(self._path, "a") as file:
            json.dump(data, file, indent=indent)
            file.write("\n")
